//
// Copyright (c) 2010-2011 StreamBase Systems, Inc. All rights reserved.
//
package com.streambase.sb.contrib.adapter.jmx;


import java.io.IOException;
import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryManagerMXBean;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.OperatingSystemMXBean;
import java.lang.management.RuntimeMXBean;
import java.lang.management.ThreadMXBean;
import java.util.ArrayList;
import java.util.List;

import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.slf4j.Logger;

import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;
import com.streambase.sb.util.Msg;

/**
 * The JMX Adapter reads JVM information from JMX beans and provides it to StreamBase applications.
 * Various outputs can be configured, depending on interest.
 * The adapter can be configured to connect to the local JVM running the application, or can
 * be given a hostname and port number to connect to. The remote JVM needs to have remote JMX enabled with
 * security disabled.
 * 
 */
public class JMXInputAdapter extends InputAdapter implements Parameterizable,
		Runnable {

	public static final long serialVersionUID = 1297275870334L;
	private Logger logger;
	// Properties
	private boolean connectLocal;
	
	private int GCPort;
	private int MMPort;
	private int MpoolPort;
	private int MemPort;
	private int ThreadPort;
	private int GeneralPort;
	private int SystemPort;
	
	private String uri;
	private int reconnectInterval;
	private int outputInterval;
	
	private boolean enableGC;
	private boolean enableMM;
	private boolean enableMpool;
	private boolean enableMem;
	private boolean enableThread;
	private boolean enableGeneral;
	private boolean enableSystem;
		
	private String displayName = "JMX Input Adapter";

	// The bean server connection object used for remote connections.
    private MBeanServerConnection  mbsConnection = null;
    
    // The collection of beans used.
    private ThreadMXBean thread = null; 
    private List<MemoryManagerMXBean> memoryManagers = null;
    private List<GarbageCollectorMXBean> garbageCollectors = null;
    private List<MemoryPoolMXBean> memoryPools = null;
    private MemoryMXBean memory = null;
    private OperatingSystemMXBean os = null;
    private RuntimeMXBean runtime = null;
    
	/**
	 * Construct the JMX adapter
	 */
	public JMXInputAdapter() {
		super();
		logger = getLogger();
		setPortHints(0, 1);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setConnectLocal(true);
		setReconnectInterval(1);
		setOutputInterval(1000);
		
	}

	/**
	 * Typecheck the parameters
	 */
	public void typecheck() throws TypecheckException {

		// Status port is always first.
		int portNum=0;
		setOutputSchema(portNum++, StatusSchemaHelper.SCHEMA);
		
		if (enableGC) {
			GCPort=portNum;
			setOutputSchema(portNum++, GarbageCollectorSchemaHelper.SCHEMA);
		}
		if (enableMpool) {
			MpoolPort=portNum;
			setOutputSchema(portNum++, MemoryPoolSchemaHelper.SCHEMA);
		}
		if (enableMM) {
			MMPort=portNum;
			setOutputSchema(portNum++, MemoryManagerSchemaHelper.SCHEMA);
		}
		if (enableMem) {
			MemPort=portNum;
			setOutputSchema(portNum++, MemorySchemaHelper.SCHEMA);
		}
		if (enableThread) {
			ThreadPort=portNum;
			setOutputSchema(portNum++, ThreadSchemaHelper.SCHEMA);
		}
		if (enableGeneral) {
			GeneralPort=portNum;
			setOutputSchema(portNum++, GeneralSchemaHelper.SCHEMA);
		}
		if (enableSystem) {
			SystemPort=portNum;
			setOutputSchema(portNum++, SystemSchemaHelper.SCHEMA);
		}
		if (!connectLocal && uri.isEmpty()) {
			throw new TypecheckException(getLocation(), "If local server not selected, you must provide a remote host/port");
		}
	}

	/**
	 * Initialize the adapter - just register the run thread
	 */
	public void init() throws StreamBaseException {
		super.init();

		// Register the object so it will be run as a thread managed by StreamBase.
		registerRunnable(this, true);
	}

	/**
	 * Shutdown adapter.
	 */
	public void shutdown() {

	}

	// Helper method to send status messages
	private void sendStatus(String type, String object, String action, String message) {
		try {
			sendOutput(0,StatusSchemaHelper.populateTuple(getOutputSchema(0).createTuple(), type, object, action, message));
		} catch (TupleException e) {
			logger.debug(getName(), e);
			logger.warn(getName(), e);
		} catch (StreamBaseException e) {
			logger.debug(getName(), e);
			logger.warn(getName(), e);
		}
	}
	
	/**
	 * The run thread will (re)connect to the JMX server and send JMX data
	 */
	public void run() {
		logger.debug(Msg.format("{0}: starting run thread", getName()));

		boolean connected=false;
		while (shouldRun()) {
			
			while (!connected && shouldRun()) {
				if (!(connected = getBeans())) {
					try {
						Thread.sleep(getReconnectInterval());
					} catch (InterruptedException e) {}
				}
			}
			
			try {
				sendJMXData();
			} catch (Throwable e) {
				connected=false;
			    thread = null; 
			    memoryManagers = null;
			    garbageCollectors = null;
			    memoryPools = null;
			    memory = null;
			    os = null;
			    runtime = null;
				logger.debug(getName(), e);
				logger.warn(Msg.format("{0}: Exception while getting JMX data: ", getName(), e));
				sendStatus(StatusSchemaHelper.StatusType.CONNECTION.toString(),
						getName(),
						StatusSchemaHelper.StatusAction.DISCONNECTED.toString(),
						e.getMessage());
			}
				
			try {
				Thread.sleep(getOutputInterval());
			} catch (InterruptedException e) {}	
		}
	}

	
	// Get the beans for all the configured outputs
	private boolean getBeans() {
		try {
			
			sendStatus(StatusSchemaHelper.StatusType.CONNECTION.toString(),
					getName(),
					StatusSchemaHelper.StatusAction.CONNECTING.toString(),
					connectLocal ? "Local JVM" : uri);
			
			// Get the bean proxy, if not local
			if (!connectLocal) {
				JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://" + getUri() + "/jmxrmi");
			    JMXConnector jmxc = JMXConnectorFactory.connect(url, null);
			    mbsConnection = jmxc.getMBeanServerConnection();
			}
			
			// Only get the beans that we are reporting one
			if (enableMM)
				memoryManagers = getMMBeans();
			
			if (enableGC)
				garbageCollectors = getGCBeans();
			
			if (enableMpool)
				memoryPools = getMPBeans();
			
			if (enableMem)
				memory = getMBeans();
			
			if (enableThread)
				thread = getThBeans();
			
			if (enableGeneral || enableSystem) {
				runtime = geRuBeans();
				os = getOSBeans();
			}
			
			sendStatus(StatusSchemaHelper.StatusType.CONNECTION.toString(),
					getName(),
					StatusSchemaHelper.StatusAction.CONNECTED.toString(),
					connectLocal ? "Local JVM" : uri);
			
		} catch (IOException e) {
			logger.debug(Msg.format("{0}", getName()), e);
			sendStatus(StatusSchemaHelper.StatusType.CONNECTION.toString(),
					getName(),
					StatusSchemaHelper.StatusAction.FAILED.toString(),
					e.getMessage());
			
			return false;
		} catch (StreamBaseException e) {
			logger.debug(Msg.format("{0}", getName()), e);
			return false;
		}
	
		return true;
	}
	
	// This set of methods will return the local bean server, or
	// reach out to the remote server via a proxy.
	private OperatingSystemMXBean getOSBeans() throws IOException {
		if (connectLocal) {
			return ManagementFactory.getOperatingSystemMXBean();
		}
	    
    	return ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
    			ManagementFactory.OPERATING_SYSTEM_MXBEAN_NAME,
    			OperatingSystemMXBean.class); 
	}
	
	private RuntimeMXBean geRuBeans() throws IOException {
		if (connectLocal) {
			return ManagementFactory.getRuntimeMXBean();
		}
	    
    	return ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
    			ManagementFactory.RUNTIME_MXBEAN_NAME,
    			RuntimeMXBean.class); 
	}
	
	private ThreadMXBean getThBeans() throws IOException {		
		if (connectLocal) {
			return ManagementFactory.getThreadMXBean();
		}
	    
    	return ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
    			ManagementFactory.THREAD_MXBEAN_NAME,
    			ThreadMXBean.class);  
	}
	
	private MemoryMXBean getMBeans() throws IOException {		
		if (connectLocal) {
			return ManagementFactory.getMemoryMXBean();
		}
	    
    	return ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
    			ManagementFactory.MEMORY_MXBEAN_NAME,
        		MemoryMXBean.class);  
	}
	
	private List<MemoryPoolMXBean> getMPBeans() throws IOException, StreamBaseException {		
		if (connectLocal) {
			return ManagementFactory.getMemoryPoolMXBeans();
		}
		
		List<MemoryPoolMXBean> mps = new ArrayList<MemoryPoolMXBean>();
	    
	    ObjectName mpName=null;
		try {
			mpName = new ObjectName(ManagementFactory.MEMORY_POOL_MXBEAN_DOMAIN_TYPE + ",*");
		} catch (MalformedObjectNameException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		} catch (NullPointerException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		}

	    for (ObjectName name : mbsConnection.queryNames(mpName, null)) {
	    	MemoryPoolMXBean mp = ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
	        		name.getCanonicalName(),
	        		MemoryPoolMXBean.class);            
	        mps.add(mp);          
	    }
		
		return mps;
	}
	
	private List<MemoryManagerMXBean> getMMBeans() throws IOException, StreamBaseException {
		
		if (connectLocal) {
			return ManagementFactory.getMemoryManagerMXBeans();
		}
		
		List<MemoryManagerMXBean> mms = new ArrayList<MemoryManagerMXBean>();
	    
	    ObjectName mmName=null;
		try {
			mmName = new ObjectName(ManagementFactory.MEMORY_MANAGER_MXBEAN_DOMAIN_TYPE + ",*");
		} catch (MalformedObjectNameException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		} catch (NullPointerException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		}

	    for (ObjectName name : mbsConnection.queryNames(mmName, null)) {
	    	MemoryManagerMXBean mm = ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
	        		name.getCanonicalName(),
	        		MemoryManagerMXBean.class);            
	        mms.add(mm);          
	    }
		
		return mms;
	}

	private List<GarbageCollectorMXBean> getGCBeans() throws IOException, StreamBaseException {
		
		if (connectLocal) {
			return ManagementFactory.getGarbageCollectorMXBeans();
		}
		
		List<GarbageCollectorMXBean> gcs = new ArrayList<GarbageCollectorMXBean>();
	    
	    ObjectName gcName=null;
		try {
			gcName = new ObjectName(ManagementFactory.GARBAGE_COLLECTOR_MXBEAN_DOMAIN_TYPE + ",*");
		} catch (MalformedObjectNameException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		} catch (NullPointerException e) {
			logger.debug(getName(), e);
			throw new StreamBaseException(e);
		}

	    for (ObjectName name : mbsConnection.queryNames(gcName, null)) {
	        GarbageCollectorMXBean gc = ManagementFactory.newPlatformMXBeanProxy(mbsConnection,
	        		name.getCanonicalName(),
	        		GarbageCollectorMXBean.class);            
	        gcs.add(gc);          
	    }
		
		return gcs;
	}
	
	// This the main method that reads the jxm data and sends it to the application
	protected void sendJMXData() {

		if (enableGC) {
			for (GarbageCollectorMXBean gbean : garbageCollectors) {
				try {
					sendOutput(GCPort, GarbageCollectorSchemaHelper.populateTuple(getOutputSchema(GCPort).createTuple(), gbean));
				} catch (TupleException e) {
					logger.debug(getName(), e);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
				}
			}
		}
			
		if (enableMpool) {
			for (MemoryPoolMXBean mpbean : memoryPools) {
				try {
					sendOutput(MpoolPort, MemoryPoolSchemaHelper.populateTuple(getOutputSchema(MpoolPort).createTuple(), mpbean));
				} catch (TupleException e) {
					logger.debug(getName(), e);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
				}
			}
		}
		
		if (enableMem) {
			try {
				sendOutput(MemPort, MemorySchemaHelper.populateTuple(getOutputSchema(MemPort).createTuple(), memory));
			} catch (TupleException e) {
				logger.debug(getName(), e);
			} catch (StreamBaseException e) {
				logger.debug(getName(), e);;
			}
		}
		
		if (enableMM) {
			for (MemoryManagerMXBean mmbean : memoryManagers) {
				try {
					sendOutput(MMPort, MemoryManagerSchemaHelper.populateTuple(getOutputSchema(MMPort).createTuple(), mmbean));
				} catch (TupleException e) {
					logger.debug(getName(), e);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
				}
			}
		}
		
		if (enableThread) {
			long [] ids = thread.getAllThreadIds();
			
			for (int i=0; i< ids.length; i++) {
				try {
					sendOutput(ThreadPort, ThreadSchemaHelper.populateTuple(getOutputSchema(ThreadPort).createTuple(), thread, ids[i]));
				} catch (TupleException e) {
					logger.debug(getName(), e);
				} catch (StreamBaseException e) {
					logger.debug(getName(), e);
				}
			}
		}
		
		if (enableGeneral) {
			try {
				sendOutput(GeneralPort, GeneralSchemaHelper.populateTuple(getOutputSchema(GeneralPort).createTuple(), os, runtime));
			} catch (TupleException e) {
				logger.debug(getName(), e);
			} catch (StreamBaseException e) {
				logger.debug(getName(), e);;
			}
		}
	}

	// Any tuple of any schema will cause a new system output to be generated.
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
		try {
			if ((os != null) && (runtime != null)) {
				sendOutput(SystemPort, SystemSchemaHelper.populateTuple(getOutputSchema(SystemPort).createTuple(), os, runtime));
			} else {
				logger.warn(Msg.format("{0}: Can't request system information, not currently connected.", getName()));
			}
		} catch (TupleException e) {
			logger.debug(getName(), e);
		} catch (StreamBaseException e) {
			logger.debug(getName(), e);;
		}
	}
	
	public PortCounts getPortCounts() {
		// There's always a status port
		int count=1;
		
		if (enableGC)
			count++;
		if (enableMpool) 
			count++;
		if (enableMem)
			count++;
		if (enableMM)
			count++;
		if (enableThread)
			count++;
		if (enableGeneral)
			count++;
		if (enableSystem)
			count++;
		
		return new PortCounts(enableSystem ? 1 : 0, count);
	}
	
	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/

	public void setConnectLocal(boolean connectLocal) {
		this.connectLocal = connectLocal;
	}
	public boolean getConnectLocal() {
		return this.connectLocal;
	}
	
	public void setUri(String uri) {
		this.uri = uri;
	}
	public String getUri() {
		return this.uri;
	}
	public boolean shouldEnableUri() {
		return !connectLocal;
	}

	public void setReconnectInterval(int secs) {
		this.reconnectInterval = secs;
	}
	public int getReconnectInterval() {
		return this.reconnectInterval;
	}
	public boolean shouldEnableReconnectInterval() {
		return !connectLocal;
	}
	
	public void setOutputInterval(int MS) {
		this.outputInterval = MS;
	}
	public int getOutputInterval() {
		return this.outputInterval;
	}
	
	public void setEnableGarbageCollector(boolean enableGC) {
		this.enableGC = enableGC;
	}
	public boolean getEnableGarbageCollector() {
		return this.enableGC;
	}
	
	public void setEnableMemoryManager(boolean enableMM) {
		this.enableMM = enableMM;
	}
	public boolean getEnableMemoryManager() {
		return this.enableMM;
	}
	
	public void setEnableMemoryPool(boolean enableMpool) {
		this.enableMpool = enableMpool;
	}
	public boolean getEnableMemoryPool() {
		return this.enableMpool;
	}
	
	public void setEnableMemory(boolean enableMem) {
		this.enableMem = enableMem;
	}
	public boolean getEnableMemory() {
		return this.enableMem;
	}

	public void setEnableThread(boolean enableThread) {
		this.enableThread = enableThread;
	}
	public boolean getEnableThread() {
		return this.enableThread;
	}
	
	public void setEnableGeneral(boolean enableGeneral) {
		this.enableGeneral = enableGeneral;
	}
	public boolean getEnableGeneral() {
		return this.enableGeneral;
	}
	
	public void setEnableSystem(boolean EnableSystem) {
		this.enableSystem = EnableSystem;
	}
	public boolean getEnableSystem() {
		return this.enableSystem;
	}
	
}
